"""*Video handling module*"""

import json
import os
from fractions import Fraction
from typing import Tuple, List, Optional

from langcodes import standardize_tag
from loguru import logger as log
import io
import sys
import tempfile
from pathlib import Path
from secrets import token_hex
from PIL import Image, ImageEnhance
import jmespath
import iscc_sdk as idk
import iscc_lib as il


__all__ = [
    "video_meta_extract",
    "video_meta_embed",
    "video_thumbnail",
    "video_mp7sig_extract",
    "video_mp7sig_extract_scenes",
    "video_features_extract",
    "video_parse_scenes",
    "video_compute_granular",
]

VIDEO_META_MAP = {
    "iscc_name": "name",
    "iscc_description": "description",
    "iscc_meta": "meta",
    "title": "name",
    "track": "name",
    "show": "name",
    "album": "name",
    "description": "description",
    "synopsis": "description",
    "comment": "description",
    "author": "creator",
    "composer": "creator",
    "artist": "creator",
    "album_artist": "creator",
    "copyright": "rights",
    "license": "license",
    "acquire": "acquire",
}


def video_meta_extract(fp):
    # type: (str | Path) -> dict
    """
    Extract video metadata using FFMPEG and FFPROBE

    :param fp: Filepath to video file
    :return: Metdata mapped to IsccMeta schema
    """
    fp = Path(fp)
    ffprobe = video_meta_extract_ffprobe(fp)
    ffmpeg = video_meta_extract_ffmpeg(fp)
    ffprobe.update(ffmpeg)
    return ffprobe


def video_meta_extract_ffprobe(fp):
    # type: (str|Path) -> dict
    """
    Extract video metadata using FFROBE

    :param fp: Filepath to video file
    :return: Metdata mapped to IsccMeta schema
    """
    fp = Path(fp)

    args = [
        "-hide_banner",
        "-loglevel",
        "fatal",
        "-find_stream_info",
        "-show_error",
        "-show_format",
        "-show_streams",
        "-show_programs",
        "-show_chapters",
        "-show_private_data",
        "-print_format",
        "json",
        "-i",
        fp,
    ]

    res = idk.run_ffprobe(args)
    metadata = json.loads(res.stdout)

    video_streams = jmespath.search("streams[?codec_type=='video']", metadata)
    n_streams = len(video_streams)
    if n_streams == 0:  # pragma: no cover
        raise ValueError("No video stream detected")
    if n_streams != 1:  # pragma: no cover
        log.warning(f"Detected {n_streams} video streams.")

    vstream = video_streams[0]

    meta = dict()

    # Duration
    duration_format = jmespath.search("format.duration", metadata)
    duration_streams = jmespath.search("streams[?codec_type=='video'].duration", metadata)
    durations = [duration_format] + duration_streams if duration_format else duration_streams
    if durations:
        duration = round(max(float(d) for d in durations))
        meta["duration"] = duration
    else:  # pragma: no cover
        log.warning("failed to extract duration")

    # Dimensions
    meta["height"] = vstream.get("height")
    meta["width"] = vstream.get("width")

    # FPS
    fps = vstream.get("r_frame_rate")
    if fps:
        fps = round(float(Fraction(fps)), 3)
        meta["fps"] = fps

    # Title
    meta["name"] = jmespath.search("format.tags.title", metadata)

    # Language
    langs = jmespath.search("streams[*].tags.language", metadata)
    lang = jmespath.search("format.tags.language", metadata)
    if lang:  # pragma: no cover
        langs.insert(0, lang)
    if langs:
        meta["language"] = standardize_tag(langs[0])

    return meta


def video_meta_extract_ffmpeg(fp):
    # type: (str|Path) -> dict
    """
    Extract metadata from video using ffmpeg.

    :param fp: Filepath to video file
    :return: Metdata mapped to IsccMeta schema
    """
    fp = Path(fp)

    args = ["-i", fp, "-movflags", "use_metadata_tags", "-f", "ffmetadata", "-"]

    result = idk.run_ffmpeg(args)
    encoding = sys.stdout.encoding or "utf-8"
    text = result.stdout.decode(encoding, errors="ignore")

    # parse metadata
    meta = dict()
    for line in text.splitlines(keepends=False):
        if not line.startswith(";FFMETA"):
            key, value = line.split("=", 1)
            if value:
                meta[key.lower()] = (
                    value.replace(r"\=", "=")
                    .replace(r"\;", ";")
                    .replace(r"\#", "#")
                    .replace(r"\;", ";")
                    .replace(r"\\n", "\n")
                    .replace(r"\\", "\\")
                )

    # map metadata
    mapped = dict()
    done = set()
    for tag, mapped_field in VIDEO_META_MAP.items():
        if mapped_field in done:
            continue
        value = meta.get(tag)
        if value:
            log.debug(f"Mapping video metadata: {tag} -> {mapped_field} -> {value}")
            mapped[mapped_field] = value
            done.add(mapped_field)
    return mapped


def video_meta_embed(fp, meta):
    # type: (str|Path, idk.IsccMeta) -> Path
    """
    Embed metadata into a copy of the video file.

    Supported fields: name, description, meta, creator, license, aquire

    :param fp: Filepath to source video file
    :param meta: Metadata to embed into video
    :return: Filepath to new video file with updated metadata
    """
    fp = Path(fp)
    write_map = {
        "name": "iscc_name",
        "description": "iscc_description",
        "meta": "iscc_meta",
        "creator": "author",
        "rights": "copyright",
        "license": "license",
        "acquire": "acquire",
    }

    # Metadata keys or values containing special characters (‘=’, ‘;’, ‘#’, ‘\’ and a newline)
    # must be escaped with a backslash ‘\’.
    escape = str.maketrans(
        {
            "=": r"\=",
            ";": r"\;",
            "#": r"\#",
            "\\": r"\\",
            "\n": r"\\n",
        }
    )

    # Prepare metadata file
    cmdf = ";FFMETADATA1\n"
    for field in write_map.keys():
        value = getattr(meta, field)
        if value:
            value = value.translate(escape)
            cmdf += f"{write_map[field]}={value}\n"

    # Create temp filepaths
    tempdir = Path(tempfile.mkdtemp())
    metafile = tempdir / "meta.txt"
    videofile = tempdir / fp.name

    # Store metadata
    with open(metafile, "wt", encoding="utf-8") as outf:
        outf.write(cmdf)

    # Embed metadata
    # See: http://ffmpeg.org/ffmpeg-formats.html#Metadata-1
    args = [
        "-i",
        fp,
        "-i",
        metafile,
        "-movflags",
        "use_metadata_tags",
        "-map_metadata",
        "1",
        "-codec",
        "copy",
        videofile,
    ]
    idk.run_ffmpeg(args)
    return videofile


def video_thumbnail(fp):
    # type: (str|Path) -> Image.Image|None
    """
    Create a thumbnail for a video.

    :param fp: Filepath to video file.
    :return: Raw PNG byte data
    :rtype: bytes
    """
    fp = Path(fp)
    size = idk.sdk_opts.image_thumbnail_size

    args = [
        "-i",
        fp,
        "-vf",
        f"thumbnail,scale={size}:-1",
        "-frames:v",
        "1",
        "-c:v",
        "png",
        "-f",
        "image2pipe",
        "-",
    ]
    try:
        result = idk.run_ffmpeg(args)
    except Exception as e:
        log.error(f"Failed video thumbnail extraction: {e}")
        return None
    img_obj = Image.open(io.BytesIO(result.stdout))
    return ImageEnhance.Sharpness(img_obj.convert("RGB")).enhance(1.4)


def video_features_extract(fp):
    # type: (str|Path) -> List[Tuple[int, ...]]
    """
    Extract video features.

    :param fp: Filepath to video file.
    :return: A sequence of frame signatures.
    """
    fp = Path(fp)
    # TODO use confidence value to improve simililarity hash.
    sig = video_mp7sig_extract(fp)

    if idk.sdk_opts.video_store_mp7sig:
        outp = fp.as_posix() + ".iscc.mp7sig"
        with open(outp, "wb") as outf:
            outf.write(sig)

    frames = idk.read_mp7_signature(sig)
    return [tuple(frame.vector.tolist()) for frame in frames]


def video_mp7sig_extract(fp):
    # type: (str|Path) -> bytes
    """Extract MPEG-7 Video Signature.

    :param fp: Filepath to video file.
    :return: raw signature data
    """
    fp = Path(fp)

    sigfile_path = Path(tempfile.mkdtemp(), token_hex(16) + ".bin")
    sigfile_path_escaped = sigfile_path.as_posix().replace(":", "\\\\:")

    # Extract MP7 Signature
    vf = f"signature=format=binary:filename={sigfile_path_escaped}"
    vf = f"fps=fps={idk.sdk_opts.video_fps}," + vf
    args = ["-i", fp, "-vf", vf, "-f", "null", "-"]
    idk.run_ffmpeg(args)

    with open(sigfile_path, "rb") as sig:
        sigdata = sig.read()
    os.remove(sigfile_path)
    return sigdata


def video_mp7sig_extract_scenes(fp, scene_limit=None):
    # type: (str|Path, int|None) -> tuple[bytes, list[float]]
    """Extract MPEG-7 Video Signature and Scenes.

    :param fp: Filepath to video file.
    :param scene_limit: Threshold value above which a scene cut is created (0.4)
    :return: tuple of raw signature data and list of scene cutpoints
    """
    fp = Path(fp)

    scene_limit = scene_limit or idk.sdk_opts.video_scene_limit

    sigfile_path = Path(tempfile.mkdtemp(), token_hex(16) + ".bin")
    sigfile_path_escaped = sigfile_path.as_posix().replace(":", "\\\\:")

    scene_path = Path(tempfile.mkdtemp(), token_hex(16) + ".cut")
    scene_path_escaped = scene_path.as_posix().replace(":", "\\\\:")

    sig_cmd = f"signature=format=binary:filename={sigfile_path_escaped}"
    sig_cmd = f"fps=fps={idk.sdk_opts.video_fps}," + sig_cmd
    scene_cmd = f"select='gte(scene,0)',metadata=print:file={scene_path_escaped}"

    args = [
        "-i",
        fp,
        "-an",
        "-sn",
        "-filter_complex",
        f"split[in1][in2];[in1]{scene_cmd}[out1];[in2]{sig_cmd}[out2]",
        "-map",
        "[out1]",
        "-f",
        "null",
        "-",
        "-map",
        "[out2]",
        "-f",
        "null",
        "-",
    ]

    idk.run_ffmpeg(args)

    with open(sigfile_path, "rb") as sig:
        sigdata = sig.read()
    os.remove(sigfile_path)

    with open(scene_path, "rt", encoding="utf-8") as scenein:
        scenetext = scenein.read()
    os.remove(scene_path)

    scenes = video_parse_scenes(scenetext, scene_limit)

    return sigdata, scenes


def video_parse_scenes(scene_text, scene_limit=None):
    # type: (str, int|None) -> List[float]
    """
    Parse scene score output from ffmpeg

    :param scene_text: Scene score output from ffmpeg
    :param scene_limit: Threshold value above which a scene cut is created (0.4)
    :return: Scene cutpoints
    """

    scene_limit = scene_limit or idk.sdk_opts.video_scene_limit

    if not scene_text.strip():
        return []

    times = []
    scores = []
    for line in scene_text.splitlines():
        if line.startswith("frame:"):
            ts = round(float(line.split()[-1].split(":")[-1]), 3)
            times.append(ts)
        if line.startswith("lavfi.scene_score"):
            scores.append(float(line.split("=")[-1]))

    cutpoints = []
    for ts, score in zip(times, scores):
        if score >= scene_limit:
            cutpoints.append(ts)

    # append last frame timestamp if not in cutpoints
    if cutpoints and cutpoints[-1] != times[-1]:
        cutpoints.append(times[-1])

    return cutpoints[1:]


def video_compute_granular(frames, scenes):
    # type: (List[idk.Frame], List[float]) -> dict
    """
    Compute video signatures for individual scenes in video.

    :param frames: List of video frames.
    :param scenes: List of video scene cutpints.
    :return: A dictionary conforming to `shema.Feature`- objects.
    """
    features, sizes, segment = [], [], []
    start_frame = 0
    for cidx, cutpoint in enumerate(scenes):
        try:
            frames = frames[start_frame:]
        except IndexError:  # pragma: no cover
            break
        for fidx, frame in enumerate(frames):
            frame_t = tuple(frame.vector.tolist())
            segment.append(frame_t)
            if frame.elapsed >= cutpoint:
                features.append(il.encode_base64(il.soft_hash_video_v0(segment, 256)))
                segment = []
                prev_cutpoint = 0 if cidx == 0 else scenes[cidx - 1]
                duration = round(cutpoint - prev_cutpoint, 3)
                sizes.append(duration)
                start_frame = fidx + 1
                break
    if not features:
        log.info("No scenes detected. Use all frames")
        segment = [tuple(frame.vector.tolist()) for frame in frames]
        features = [il.encode_base64(il.soft_hash_video_v0(segment, bits=256))]
        sizes = [round(float(frames[-1].elapsed), 3)]

    return dict(maintype="content", subtype="video", version=0, simprints=features, sizes=sizes)



# -*- coding: utf-8 -*-


class IsccError(Exception):
    pass


class IsccExtractionError(IsccError):
    pass


class IsccUnsupportedMediatype(IsccError):
    pass


class EnvironmentError(IsccError):
    pass



"""*SDK main top-level functions*."""

from typing import Any

from loguru import logger as log
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from PIL import Image
import iscc_lib as il
import iscc_sdk as idk


__all__ = [
    "code_iscc",
    "code_meta",
    "code_content",
    "code_text",
    "code_text_semantic",
    "code_image",
    "code_image_semantic",
    "code_audio",
    "code_video",
    "code_data",
    "code_instance",
    "code_sum",
]


def code_iscc(fp, name=None, description=None, meta=None, **options):
    # type: (str | Path, str | None, str | None, str | dict | None, Any) -> idk.IsccMeta
    """
    Generate a complete ISCC-CODE for the given file.

    This function creates a full ISCC-CODE by combining Meta, Content, Data, and Instance Codes.
    It automatically detects the media type and processes the file accordingly.

    The function performs the following steps:
    1. Reads the file and determines its media type.
    2. Generates Data & Instance Codes for all file types using the `code_sum` function.
    3. For supported media types, generates Content Code and optional Semantic Code.
    4. If enabled, generates Meta-Code from embedded or provided metadata.
    5. Combines all generated code units into a single ISCC-CODE.
    6. Merges metadata from all ISCC units.

    ISCC-CODE is a composite identifier that consists of multiple ISCC-UNITs, each serving a
    specific purpose:
    - Meta-Code: Based on normalized metadata (title, description)
    - Semantic-Code: Based on semantic features (experimental, requires additional packages)
    - Content-Code: Based on the content features (text, image, audio, video)
    - Data-Code: Based on the raw binary data (similarity preserving hash)
    - Instance-Code: Based on the exact binary data (cryptographic hash)

    Note:
    - The behavior can be customized through the `sdk_opts` settings. For example, setting
      `fallback` to True will allow processing of unsupported media types in a
      fallback mode instead of raising an exception.
    - For processing container files (like EPUB with embedded files), set `process_container`
      to True to extract and process contained files.

    :param fp: Path object or str representing the filepath of the file to process.
    :param name: Optional name to override extracted metadata.
    :param description: Optional description to override extracted metadata.
    :param meta: Optional metadata (dict or Data-URL as string) to override extracted metadata.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **fallback** - Process unsupported media types. Default: False;
        **add_units** - Include ISCC-UNITS in metadata. Default: False;
        **create_meta** - Create Meta-Code. Default: True;
        **wide** - Enable wide mode for ISCC-SUM with Data & Instance codes only. Default: False;
        **experimental** - Enable experimental semantic codes. Default: False;
        **process_container** - Process container files and extract contained files. Default: False;
        **granular** - Generate additional granular fingerprints. Default: False
    :return: IsccMeta object with complete ISCC-CODE and merged metadata from all ISCC-UNITs.
    :raises idk.IsccUnsupportedMediatype:
        If the media type is not supported. By default, the function will raise this exception for
        unsupported media types, as sdk_opts.fallback is False by default.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    # Initialize collectors
    iscc_meta = dict(filename=fp.name)

    with open(fp, "rb") as infile:
        data = infile.read(4096)

    mediatype = idk.mediatype_guess(data, file_name=fp.name)
    iscc_meta["mediatype"] = mediatype

    try:
        mode = idk.mediatype_to_mode(mediatype)
        log.debug(f"Processing {fp.name} - media type: {mediatype} - processing mode: {mode}")
    except idk.IsccUnsupportedMediatype:
        if not opts.fallback:
            raise
        mode = None
        log.warning(f"Processing {fp.name} - media type: {mediatype} - processing mode: {mode}")

    iscc_meta["mediatype"] = mediatype
    if mode:
        schema_org_map = {
            "text": "TextDigitalDocument",
            "image": "ImageObject",
            "audio": "AudioObject",
            "video": "VideoObject",
        }
        type_ = schema_org_map.get(mode)
        if type_:
            iscc_meta["@type"] = type_
        iscc_meta["mode"] = mode

    # Generate Data & Instance Codes
    iscc_sum = code_sum(fp, **options)

    # Generate Content & optional Semantic Codes
    cc = None
    cs = None
    if mode == "image":
        cc = code_image(fp, **options)
        if idk.is_installed("iscc_sci") and opts.experimental:  # pragma: nocover
            cs = code_image_semantic(fp)
    elif mode == "audio":
        cc = code_audio(fp, **options)
    elif mode == "video":
        cc = code_video(fp, **options)
    elif mode == "text":
        text = idk.text_extract(fp)
        text = il.text_clean(text)
        cc = code_text(fp, text, **options)
        if idk.is_installed("iscc_sct") and opts.experimental:  # pragma: nocover
            cs = code_text_semantic(fp, text)  # Don´t pass incopatible options here!

    # Generate Meta-Code
    meta = code_meta(fp, name, description, meta, **options) if opts.create_meta and mode else None

    # Collect Metadata
    iscc_meta.update(iscc_sum.dict())
    if cs:  # pragma: nocover
        iscc_meta.update(cs.dict())
    if cc:
        iscc_meta.update(cc.dict())
    if meta:
        iscc_meta.update(meta.dict())

    # Add ISCC-UNITS
    iscc_units = []
    if meta:
        iscc_units.append(meta.iscc)
    if cs:  # pragma: nocover
        iscc_units.append(cs.iscc)
    if cc:
        iscc_units.append(cc.iscc)

    if hasattr(iscc_sum, "units") and iscc_sum.units:
        iscc_units.extend(iscc_sum.units)
    else:
        iscc_units.extend(f"ISCC:{u}" for u in il.iscc_decompose(iscc_sum.iscc))

    if opts.add_units:
        iscc_meta["units"] = iscc_units

    # Add granular features
    if opts.granular:
        features = []
        if hasattr(cs, "features") and cs.features:  # pragma: nocover
            features.append(cs.features[0])
        if hasattr(cc, "features") and cc.features:
            features.append(cc.features[0])
        iscc_meta["features"] = features

    # Compose ISCC-CODE
    iscc_code = il.gen_iscc_code_v0(iscc_units, wide=opts.wide)
    iscc_meta.update(iscc_code)
    iscc_meta["generator"] = f"iscc-sdk - v{idk.__version__}"

    result = idk.IsccMeta.model_construct(**iscc_meta)

    if opts.process_container:
        parts = idk.process_container(fp, **options)
        if parts:
            result.parts = parts

    return result


def code_iscc_mt(fp, name=None, description=None, meta=None, **options):  # pragma: no cover
    # type: (str|Path, str|None, str|None, str|dict|None, Any) -> idk.IsccMeta
    """
    Generate a complete ISCC-CODE for the given file using multithreading.

    This function creates a full ISCC-CODE by combining Meta, Content, Data, and Instance Codes.
    It automatically detects the media type and processes the file accordingly.

    The function performs the following steps:
    1. Reads the file and determines its MIME type.
    2. Generates Instance and Data Codes for all file types.
    3. For supported media types, generates Content and Meta Codes.
    4. Combines all generated codes into a single ISCC-CODE.
    5. Merges metadata from all ISCC units.

    Note:

    - This function uses multithreading to improve performance.
    - The behavior can be customized through the `sdk_opts` settings. For example, setting
      `fallback` to True will allow processing of unsupported media types in a
      fallback mode instead of raising an exception.

    :param fp: str or Path object representing the filepath of the file to process.
    :param name: Optional name to override extracted metadata.
    :param description: Optional description to override extracted metadata.
    :param meta: Optional metadata (dict or Data-URL as string) to override extracted metadata.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **fallback** - Process unsupported media types. Default: False;
        **add_units** - Include ISCC-UNITS in metadata. Default: False;
        **create_meta** - Create Meta-Code unit from embedded metadata. Default: True;
        **wide** - Enable wide mode for ISCC-SUM with Data & Instance codes only. Default: False;
        **experimental** - Enable experimental semantic codes. Default: False
    :return: IsccMeta object with complete ISCC-CODE and merged metadata from all ISCC-UNITs.
    :raises idk.IsccUnsupportedMediatype:
        If the media type is not supported. By default, the function will raise this exception for
        unsupported media types, as sdk_opts.fallback is False by default.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    iscc_meta = dict(filename=fp.name)

    # Track list properties for custom merging
    iscc_units = []
    iscc_features = []

    with open(fp, "rb") as infile:
        data = infile.read(4096)

    mime = idk.mediatype_guess(data, file_name=fp.name)
    iscc_meta["mediatype"] = mime

    try:
        mode = idk.mediatype_to_mode(mime)
        log.debug(f"Processing {fp.name} - media type: {mime} - processing mode: {mode}")
    except idk.IsccUnsupportedMediatype:
        if not opts.fallback:
            raise
        mode = None
        log.warning(f"Processing {fp.name} - media type: {mime} - processing mode: {mode}")

    with ThreadPoolExecutor() as executor:
        # Always process instance and data-codes first
        sum_future = executor.submit(code_sum, fp, **options)

        if mode is not None:
            # Process content and meta for supported media types
            content_future = executor.submit(code_content, fp, **options)
            if opts.create_meta:
                meta_future = executor.submit(code_meta, fp, name, description, meta, **options)
                meta = meta_future.result()
                iscc_units.append(meta.iscc)

            # Optional semantic codes
            if mode == "image" and idk.is_installed("iscc_sci") and opts.experimental:
                content_semantic_future = executor.submit(code_image_semantic, fp)
            elif mode == "text" and idk.is_installed("iscc_sct") and opts.experimental:
                content_semantic_future = executor.submit(code_text_semantic, fp)
            else:
                content_semantic_future = None

            if content_semantic_future:
                content_semantic = content_semantic_future.result()
                iscc_units.append(content_semantic.iscc)
                if content_semantic.features:
                    iscc_features.append(content_semantic.features[0])
                iscc_meta.update(content_semantic.dict(exclude={"features"}))

            content = content_future.result()
            iscc_units.append(content.iscc)
            if content.features:
                iscc_features.append(content.features[0])
            iscc_meta.update(content.dict(exclude={"features"}))
            if opts.create_meta:
                iscc_meta.update(meta.dict())

        # Wait for instance and data to complete
        sum_result = sum_future.result()
        if hasattr(sum_result, "units") and sum_result.units:
            iscc_units.extend(sum_result.units)
        else:
            iscc_units.extend(f"ISCC:{u}" for u in il.iscc_decompose(sum_result.iscc))
        iscc_meta.update(sum_result.dict())

    if opts.add_units:
        iscc_meta["units"] = iscc_units
    if iscc_features:
        iscc_meta["features"] = iscc_features

    # Compose ISCC-CODE
    iscc_code = il.gen_iscc_code_v0(iscc_units, wide=opts.wide)
    iscc_meta.update(iscc_code)
    iscc_meta["generator"] = f"iscc-sdk - v{idk.__version__}"

    result = idk.IsccMeta.model_construct(**iscc_meta)

    if opts.process_container:
        parts = idk.process_container(fp, **options)
        if parts:
            result.parts = parts

    return result


def code_meta(fp, name=None, description=None, meta=None, **options):
    # type: (str|Path, str|None, str|None, str|dict|None, Any) -> idk.IsccMeta
    """
    Generate Meta-Code for digital asset.

    Creates an ISCC Meta-Code based on normalized metadata extracted from the file.
    If no name is found in metadata, the filename will be used instead.

    :param fp: Filepath used for Meta-Code creation.
    :param name: Optional name to override extracted metadata.
    :param description: Optional description to override extracted metadata.
    :param meta: Optional metadata (Data-URL as string or dict) to override extracted metadata.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **bits** - Bit-length of the generated Meta-Code UNIT. Default: 64
    :return: ISCC metadata including Meta-Code and extracted metadata fields.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    meta_dict = dict()

    if opts.extract_meta:
        meta_dict = idk.extract_metadata(fp).dict()

    # Override with provided parameters if they exist
    if name is not None:
        meta_dict["name"] = name
    if description is not None:
        meta_dict["description"] = description
    if meta is not None:
        meta_dict["meta"] = meta

    # Pre-Check if we have a name after normalization, else use filename.
    name_val = meta_dict.get("name")
    name_val = "" if name_val is None else name_val
    name_val = il.text_clean(name_val)
    name_val = il.text_remove_newlines(name_val)
    name_val = il.text_trim(name_val, il.core_opts.meta_trim_name)
    if not name_val:
        meta_dict["name"] = idk.text_name_from_uri(fp)
        log.warning(f"Acquired Meta-Code `name` from filename: {meta_dict['name']}")

    metacode = il.gen_meta_code_v0(
        name=meta_dict.get("name"),
        description=meta_dict.get("description"),
        meta=meta_dict.get("meta"),
        bits=opts.bits,
    )

    meta_dict.update(metacode)
    return idk.IsccMeta.model_construct(**meta_dict)


def code_content(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Detect mediatype and create corresponding Content-Code.

    Analyzes the file to determine its media type and routes the processing to the
    appropriate specialized function (code_text, code_image, code_audio, or code_video).

    :param fp: Filepath
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **create_thumb** - Whether to create a thumbnail. Default: True
    :return: Content-Code wrapped in ISCC metadata.
    :raises idk.IsccUnsupportedMediatype: If the media type is not supported.
    """
    fp = Path(fp)
    schema_org_map = {
        "text": "TextDigitalDocument",
        "image": "ImageObject",
        "audio": "AudioObject",
        "video": "VideoObject",
    }

    mediatype, mode = idk.mediatype_and_mode(fp)

    if mode == "image":
        cc = code_image(fp, **options)
    elif mode == "audio":
        cc = code_audio(fp, **options)
    elif mode == "video":
        cc = code_video(fp, **options)
    elif mode == "text":
        cc = code_text(fp, **options)
    else:  # pragma nocover
        raise idk.IsccUnsupportedMediatype(mediatype)

    cc.mediatype = mediatype
    cc.mode = mode
    cc.type_ = schema_org_map.get(mode)

    return cc


def code_text(fp, text=None, **options):
    # type: (str|Path, str|None, Any) -> idk.IsccMeta
    """
    Generate Content-Code Text.

    Creates a Text-Code by extracting and processing text content from document files.
    Can optionally extract metadata and create a thumbnail representation of the text.

    :param fp: Filepath used for Text-Code creation.
    :param text: Optional cleaned text. If provided, the function will skip text extraction.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **create_thumb** - Whether to create a thumbnail. Default: True;
        **bits** - Bit-length of the generated Text-Code UNIT. Default: 64;
        **granular** - Whether to generate additional granular fingerprints. Default: False
    :return: ISCC metadata including Text-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.text_meta_extract(fp)

    if opts.create_thumb:
        thumbnail_img = idk.text_thumbnail(fp)
        if thumbnail_img:
            thumbnail_durl = idk.image_to_data_url(thumbnail_img)
            meta["thumbnail"] = thumbnail_durl

    if text is None:
        text = idk.text_extract(fp)
        text = il.text_clean(text)

    code = il.gen_text_code_v0(text, bits=opts.bits)
    meta.update(code)
    if opts.granular:
        features = idk.text_features(text)
        meta["features"] = [features]
    if opts.text_keep:
        meta["text"] = text
    return idk.IsccMeta.model_construct(**meta)


def code_text_semantic(fp, text=None, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Semantic-Code Text. (Requires iscc-sct to be installed)
    :param fp: Filepath used for semantic Text-Code creation.
    :param text: Optional cleaned text. If provided, the function will skip text extraction.
    :raises idk.EnvironmentError: If iscc-sct is not installed.
    """
    if not idk.is_installed("iscc_sct"):
        raise idk.EnvironmentError(
            "Semantic-Code Text requires `iscc-sct` package to be installed."
        )
    else:  # pragma: nocover
        import iscc_sct

        fp = Path(fp)
        if text is None:
            text = idk.text_extract(fp)
            text = il.text_clean(text)

        result = iscc_sct.gen_text_code_semantic(text, **options)
        return idk.IsccMeta.model_construct(**result)


def code_image(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Image.

    Creates an Image-Code by normalizing and processing the visual content of image files.
    The image is normalized according to SDK options (transparency handling, border trimming, ...).

    :param fp: Filepath used for Image-Code creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **create_thumb** - Whether to create a thumbnail. Default: True;
        **bits** - Bit-length of the generated Image-Code UNIT. Default: 64
    :return: ISCC metadata including Image-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.image_meta_extract(fp)
    if opts.create_thumb:
        thumbnail_img = idk.image_thumbnail(fp)
        thumbnail_durl = idk.image_to_data_url(thumbnail_img)
        meta["thumbnail"] = thumbnail_durl

    pixels = idk.image_normalize(Image.open(fp))
    code_obj = il.gen_image_code_v0(pixels, bits=opts.bits)
    meta.update(code_obj)

    return idk.IsccMeta.model_construct(**meta)


def code_image_semantic(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Semantic-Code Image. (Requires iscc-sci to be installed)
    :param fp: Filepath used for semantic Image-Code creation.
    :raises idk.EnvironmentError: If iscc-sci is not installed.
    """
    if not idk.is_installed("iscc_sci"):
        raise idk.EnvironmentError(
            "Semantic-Code Image requires `iscc-sci` package to be installed."
        )
    else:  # pragma: nocover
        import iscc_sci

        fp = Path(fp)
        meta = iscc_sci.code_image_semantic(fp, **options)
        return idk.IsccMeta.model_construct(**meta)


def code_audio(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Audio.

    Creates an Audio-Code by extracting acoustic fingerprints from audio files.
    Uses chromaprint/fpcalc to generate audio features for similarity matching.

    :param fp: Filepath used for Audio-Code creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **create_thumb** - Whether to create a thumbnail. Default: True;
        **bits** - Bit-length of the generated Audio-Code UNIT. Default: 64
    :return: ISCC metadata including Audio-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.audio_meta_extract(fp)
    if opts.create_thumb:
        thumbnail_img = idk.audio_thumbnail(fp)
        if thumbnail_img:
            thumbnail_durl = idk.image_to_data_url(thumbnail_img)
            meta["thumbnail"] = thumbnail_durl

    features = idk.audio_features_extract(fp)
    code_obj = il.gen_audio_code_v0(features["fingerprint"], bits=opts.bits)
    meta.update(code_obj)

    return idk.IsccMeta.model_construct(**meta)


def code_video(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Video.

    Creates a Video-Code by extracting and processing visual features from video frames.
    Uses MPEG-7 signature tools to extract frame-based features and optionally detect scene changes.

    :param fp: Filepath used for Video-Code creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **extract_meta** - Whether to extract metadata. Default: True;
        **create_thumb** - Whether to create a thumbnail. Default: True;
        **granular** - Generate additional fingerprints based on scenes. Default: False;
        **video_store_mp7sig** - Whether to store extracted MP7 Video signature file. Default: False;
        **bits** - Bit-length of the generated Video-Code UNIT. Default: 64
    :return: ISCC metadata including Video-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.video_meta_extract(fp)

    if opts.create_thumb:
        thumbnail_image = idk.video_thumbnail(fp)
        if thumbnail_image is not None:
            thumbnail_durl = idk.image_to_data_url(thumbnail_image)
            meta["thumbnail"] = thumbnail_durl

    sig, scenes = None, []
    if opts.granular:
        sig, scenes = idk.video_mp7sig_extract_scenes(fp)
    else:
        sig = idk.video_mp7sig_extract(fp)

    if opts.video_store_mp7sig:
        outp = fp.with_suffix(".iscc.mp7sig")
        with open(outp, "wb") as outf:
            outf.write(sig)

    frames = idk.read_mp7_signature(sig)
    features = [tuple(frame.vector.tolist()) for frame in frames]

    code_obj = il.gen_video_code_v0(features, bits=opts.bits)
    meta.update(code_obj)

    if opts.granular:
        granular = idk.video_compute_granular(frames, scenes)
        meta["features"] = [granular]

    return idk.IsccMeta.model_construct(**meta)


def code_data(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Create ISCC Data-Code.

    The Data-Code is a similarity preserving hash of the raw input data that allows for
    detection of similar binary data regardless of file format or metadata differences.

    :param fp: Filepath used for Data-Code creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **bits** - Bit-length of the generated Data-Code UNIT. Default: 64
    :return: ISCC metadata including Data-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    with open(fp, "rb") as stream:
        result = il.gen_data_code_v0(stream, bits=opts.bits)

    return idk.IsccMeta.model_construct(**result)


def code_instance(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Create ISCC Instance-Code.

    The Instance-Code is a cryptographic hash (blake3) of the input data.
    Its purpose is to serve as a checksum that detects even minimal changes
    to the data of the referenced media asset. For cryptographically secure integrity
    checking, a full 256-bit multihash is provided with the `datahash` field.

    :param fp: Filepath used for Instance-Code creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **bits** - Bit-length of the generated Instance-Code UNIT. Default: 64
    :return: ISCC metadata including Instance-Code, datahash, and filesize.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    with open(fp, "rb") as stream:
        result = il.gen_instance_code_v0(stream, bits=opts.bits)

    return idk.IsccMeta.model_construct(**result)


def code_sum(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Create an ISCC-CODE with Data- and Instance-Code UNITs in a single pass.

    :param fp: Filepath used for ISCC-CODE Sum creation.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **bits** - Bit-length for Data-Code body. Default: 64;
        **wide** - Whether to use wide or narrow ISCC-CODE (64-bit or 128-bit UNITs);
        **add_units** - Include individual ISCC-UNITs in result. Default: False
    :return: ISCC metadata.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    result = il.gen_sum_code_v0(fp, bits=opts.bits, wide=opts.wide, add_units=opts.add_units)
    meta = {
        "iscc": result["iscc"],
        "datahash": result["datahash"],
        "filesize": result["filesize"],
    }
    if result.get("units"):
        meta["units"] = result["units"]

    return idk.IsccMeta.model_construct(**meta)



import time
import shutil
import tempfile
from functools import cache
from pathlib import Path
from loguru import logger as log


__all__ = [
    "TempFile",
    "timer",
    "is_installed",
]


class TempFile:
    def __init__(self, original_path):
        # type: (str|Path) -> None
        self.original_path = Path(original_path)
        self.temp_dir: Path | None = None

    def __enter__(self):
        # type: () -> Path
        self.temp_dir = Path(tempfile.mkdtemp())
        temp_filename = self.temp_dir / self.original_path.name
        shutil.copy2(self.original_path, temp_filename)
        return temp_filename

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.temp_dir)


class timer:
    def __init__(self, message: str):
        self.message = message

    def __enter__(self):
        # Record the start time
        self.start_time = time.perf_counter()

    def __exit__(self, exc_type, exc_value, traceback):
        # Calculate the elapsed time
        elapsed_time = time.perf_counter() - self.start_time
        # Log the message with the elapsed time
        log.debug(f"{self.message} {elapsed_time:.4f} seconds")


@cache
def is_installed(package_name):
    # type: (str) -> bool
    """
    Check if a Python package is installed.

    :param str package_name: The name of the package to check
    :return: True if the package is installed, False otherwise
    :rtype: bool
    """
    try:
        __import__(package_name)
        return True
    except ImportError:
        return False



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from fractions import Fraction
from functools import lru_cache
from typing import Tuple, List
from bitarray import bitarray
from bitarray.util import ba2int
import numpy as np


__all__ = [
    "read_mp7_signature",
    "Frame",
]


SIGELEM_SIZE = 380


@dataclass
class Frame:
    """Represents an MP7 Frame Signature."""

    vector: np.ndarray  # 380 dimensional vector, range: 0..2
    elapsed: Fraction  # time elapsed since start of video
    confidence: int  # signature confidence, range: 0..255


@lru_cache
def calc_byte_to_bit3():
    # type: () -> np.ndarray
    """
    Build lookup table.

    :return: table to convert a 8bit value into five three-bit-values
    :rtype: np.ndarray
    """
    table_3_bit = np.zeros((256, 5), dtype=np.uint8)
    for i in range(256):
        div3 = 3 * 3 * 3 * 3
        for iii in range(0, 5):
            table_3_bit[i, iii] = (i // div3) % 3
            div3 //= 3
    return table_3_bit


def pop_bits(data_bits, pos, bits=32):
    # type: (bitarray, int, int) -> Tuple[int, int]
    """
    Take out 0/1 values and pack them again to an unsigned integer.

    :param bitarray data_bits: 0/1 data
    :param int pos: position in 0/1 data
    :param int bits: number of bits (default 32)
    :return: value, new position
    :rtype: Tuple[int, int]
    """
    chunk = data_bits[pos : pos + bits]
    value = ba2int(chunk, signed=False)
    pos += bits
    return value, pos


def read_mp7_signature(byte_data):
    # type: (bytes) -> List[Frame]
    """
    Decode binary MP7 video signature.

    :param bytes byte_data: Raw MP7 video signature (as extracted by ffmpeg)
    :return: List of Frame Signatures
    :rtype: List[Frame]
    """
    table_3_bit = calc_byte_to_bit3()
    data_bits = bitarray()
    data_bits.frombytes(byte_data)
    pos = 0
    pos += 129
    num_of_frames, pos = pop_bits(data_bits, pos)
    media_time_unit, pos = pop_bits(data_bits, pos, 16)
    pos += 1 + 32 + 32
    num_of_segments, pos = pop_bits(data_bits, pos)
    pos += num_of_segments * (4 * 32 + 1 + 5 * 243)
    pos += 1
    frame_sigs_v = []
    frame_sigs_c = []
    frame_sigs_e = []
    frame_sigs_tu = []
    for i in range(num_of_frames):
        pos += 1
        raw_media_time, pos = pop_bits(data_bits, pos)
        frame_confidence, pos = pop_bits(data_bits, pos, 8)
        pos += 5 * 8
        vec = np.zeros((SIGELEM_SIZE,), dtype=np.uint8)
        p = 0
        for ii in range(SIGELEM_SIZE // 5):
            dat, pos = pop_bits(data_bits, pos, 8)
            vec[p : p + 5] = table_3_bit[dat]
            p += 5
        frame_sigs_v.append(vec)
        frame_sigs_e.append(raw_media_time)
        frame_sigs_c.append(frame_confidence)
        frame_sigs_tu.append(media_time_unit)

    fsigs = []
    r = (frame_sigs_v, frame_sigs_e, frame_sigs_c, frame_sigs_tu)
    for v, e, c, tu in zip(*r):
        fsigs.append(Frame(vector=v, elapsed=Fraction(e, tu), confidence=c))
    return fsigs



"""Pure Python IPFS CIDv1 computation (Kubo-compatible)."""

import hashlib
from base64 import b32encode
from pathlib import Path


__all__ = [
    "ipfs_cidv1",
    "ipfs_cidv1_base16",
]

# Kubo defaults for CIDv1
CHUNK_SIZE = 262144  # 256 KB fixed-size chunker
MAX_LINKS = 174  # DefaultLinksPerBlock
SHA2_256 = 0x12
DIGEST_LENGTH = 0x20
CODEC_RAW = 0x55
CODEC_DAG_PB = 0x70
UNIXFS_FILE = 2
UNIXFS_DIRECTORY = 1


def _varint_encode(n):
    """Encode unsigned integer as LEB128 varint."""
    buf = bytearray()
    while n > 0x7F:
        buf.append((n & 0x7F) | 0x80)
        n >>= 7
    buf.append(n & 0x7F)
    return bytes(buf)


def _sha256_digest(data):
    """Compute raw SHA-256 digest."""
    return hashlib.sha256(data).digest()


def _make_multihash(digest):
    """Prepend sha2-256 multihash header to digest."""
    return _varint_encode(SHA2_256) + _varint_encode(DIGEST_LENGTH) + digest


def _cid_bytes(codec, data):
    """Build CIDv1 binary from codec and data to hash."""
    digest = _sha256_digest(data)
    return _varint_encode(1) + _varint_encode(codec) + _make_multihash(digest)


def _cid_base32(cid_bytes):
    """Encode CID bytes as base32lower with 'b' prefix (no padding)."""
    return "b" + b32encode(cid_bytes).decode("ascii").lower().rstrip("=")


def _cid_base16(cid_bytes):
    """Encode CID bytes as base16 (hex) with 'f' prefix."""
    return "f" + cid_bytes.hex()


def _encode_varint_field(field_num, value):
    """Encode a protobuf varint field (wire type 0)."""
    return _varint_encode(field_num << 3 | 0) + _varint_encode(value)


def _encode_bytes_field(field_num, data):
    """Encode a protobuf length-delimited field (wire type 2)."""
    return _varint_encode(field_num << 3 | 2) + _varint_encode(len(data)) + data


def _encode_unixfs_data(type_val, filesize=None, blocksizes=None):
    """Encode UnixFS Data protobuf message."""
    buf = _encode_varint_field(1, type_val)
    if filesize is not None:
        buf += _encode_varint_field(3, filesize)
    if blocksizes:
        for bs in blocksizes:
            buf += _encode_varint_field(4, bs)
    return buf


def _encode_pb_link(cid_bytes, tsize, name=""):
    """Encode a dag-pb PBLink protobuf message.

    Kubo always serializes the Name field (even as empty string for file links).
    """
    buf = _encode_bytes_field(1, cid_bytes)
    buf += _encode_bytes_field(2, name.encode("utf-8"))
    buf += _encode_varint_field(3, tsize)
    return buf


def _encode_pb_node(links_bytes, unixfs_data):
    """Encode a dag-pb PBNode (Links before Data per spec)."""
    buf = b""
    for link in links_bytes:
        buf += _encode_bytes_field(2, link)
    buf += _encode_bytes_field(1, unixfs_data)
    return buf


def _build_balanced_dag(leaves):
    """Build balanced DAG from leaf tuples, return (cid_bytes, tsize, data_size)."""
    while len(leaves) > 1:
        next_level = []
        for i in range(0, len(leaves), MAX_LINKS):
            batch = leaves[i : i + MAX_LINKS]
            data_size = sum(leaf[2] for leaf in batch)
            blocksizes = [leaf[2] for leaf in batch]
            unixfs = _encode_unixfs_data(UNIXFS_FILE, filesize=data_size, blocksizes=blocksizes)
            links = [_encode_pb_link(leaf[0], leaf[1]) for leaf in batch]
            node_bytes = _encode_pb_node(links, unixfs)
            cid = _cid_bytes(CODEC_DAG_PB, node_bytes)
            tsize = len(node_bytes) + sum(leaf[1] for leaf in batch)
            next_level.append((cid, tsize, data_size))
        leaves = next_level
    return leaves[0]


def _process_file(fp):
    """Process file into CID, returning (cid_bytes, tsize, data_size)."""
    fp = Path(fp)
    leaves = []
    with open(fp, "rb") as f:
        while True:
            chunk = f.read(CHUNK_SIZE)
            if not chunk:
                break
            cid = _cid_bytes(CODEC_RAW, chunk)
            leaves.append((cid, len(chunk), len(chunk)))

    if not leaves:
        # Empty file: single empty raw leaf
        cid = _cid_bytes(CODEC_RAW, b"")
        return cid, 0, 0

    if len(leaves) == 1:
        return leaves[0]

    return _build_balanced_dag(leaves)


def ipfs_cidv1(fp, wrap=False):
    # type: (str|Path, bool) -> str
    """Compute IPFS CIDv1 for file (Kubo-compatible, pure Python).

    :param fp: Filepath to hash.
    :param wrap: Wrap file in a directory and append filename to CID path.
    :return: IPFS CIDv1 base32lower string.
    """
    fp = Path(fp)
    file_cid, file_tsize, _ = _process_file(fp)

    if not wrap:
        return _cid_base32(file_cid)

    # Directory wrapping: create dag-pb directory node with one named link
    unixfs = _encode_unixfs_data(UNIXFS_DIRECTORY)
    link = _encode_pb_link(file_cid, file_tsize, name=fp.name)
    dir_node = _encode_pb_node([link], unixfs)
    dir_cid = _cid_bytes(CODEC_DAG_PB, dir_node)
    return _cid_base32(dir_cid) + f"/{fp.name}"


def ipfs_cidv1_base16(fp):
    # type: (str|Path) -> str
    """Compute IPFS CIDv1 with base16 encoding (Kubo-compatible, pure Python).

    :param fp: Filepath to hash.
    :return: IPFS CIDv1 base16 (hex) string.
    """
    file_cid, _, _ = _process_file(fp)
    return _cid_base16(file_cid)



"""Monkey Patches"""

import ebookmeta
from ebookmeta import Epub2, _get_ebook


def set_author_list(self, author_list):  # pragma: no cover
    """Don´t set role on creator"""
    node_list = self._get_all('opf:metadata/dc:creator[@opf:role="aut" or not(@opf:role)]')
    for node in node_list:
        node.getparent().remove(node)
    meta_node = self._get("opf:metadata")
    for author in author_list:
        if author:
            node = self._sub_element(meta_node, "dc:creator")
            node.text = author


Epub2.set_author_list = set_author_list


def set_description(self, description):  # pragma: no cover
    node = self._get("opf:metadata/dc:description")
    if node is None:
        meta_node = self._get("opf:metadata")
        node = self._sub_element(meta_node, "dc:description")
    node.text = description


Epub2.set_description = set_description


def set_metadata(file, meta):  # pragma: no cover
    ebook = _get_ebook(file)

    ebook.set_title(meta.title)
    ebook.set_description(meta.description)
    ebook.set_author_list(meta.author_list)
    ebook.set_series(meta.series)
    ebook.set_series_index(meta.series_index)
    ebook.set_lang(meta.lang)
    ebook.set_tag_list(meta.tag_list)
    ebook.set_translator_list(meta.translator_list)
    ebook.set_cover_data(meta.cover_file_name, meta.cover_media_type, meta.cover_image_data)

    # Set publish info for FB2
    if meta.format == "fb2":
        ebook.set_publish_title(meta.publish_info.title)
        ebook.set_publish_publisher(meta.publish_info.publisher)
        ebook.set_publish_city(meta.publish_info.city)
        ebook.set_publish_year(meta.publish_info.year)
        ebook.set_publish_series(meta.publish_info.series)
        ebook.set_publish_series_index(meta.publish_info.series_index)
        ebook.set_publish_isbn(meta.publish_info.isbn)

    ebook.save()


ebookmeta.set_metadata = set_metadata



"""*PDF handling module*."""

import shutil
import tempfile
from pathlib import Path

from PIL import Image, ImageEnhance
import fitz
import iscc_sdk as idk
from pdf_oxide import PdfDocument


__all__ = [
    "pdf_thumbnail",
    "pdf_meta_embed",
    "pdf_text_extract",
]


def pdf_text_extract(fp):
    # type: (str|Path) -> str
    """Extract PDF text with reading order reconstruction via XY-Cut algorithm."""
    fp = Path(fp)
    return PdfDocument(str(fp)).to_plain_text_all(
        preserve_layout=True,
        detect_headings=True,
        include_images=False,
    )


def pdf_thumbnail(fp):
    # type: (str|Path) -> Image.Image
    """
    Create a thumbnail from PDF document.

    :param fp: Filepath to PDF document.
    :return: Thumbnail image as PIL Image object
    """
    fp = Path(fp)
    with fitz.Document(fp) as doc:
        page = doc.load_page(0)
        pix = page.get_pixmap()
        mode = "RGBA" if pix.alpha else "RGB"
        img = Image.frombytes(mode, [pix.width, pix.height], pix.samples)
        size = idk.sdk_opts.image_thumbnail_size
        img.thumbnail((size, size), resample=idk.LANCZOS)
        return ImageEnhance.Sharpness(img.convert("RGB")).enhance(1.4)


def pdf_meta_embed(fp, meta):
    # type: (str|Path, idk.IsccMeta) -> Path
    """
    Embed metadata into a copy of the PDF file.

    :param fp: Filepath to source PDF file
    :param meta: Metadata to embed into PDF
    :return: Filepath to the new PDF file with updated metadata
    """
    fp = Path(fp)
    tempdir = tempfile.mkdtemp()
    temppdf = shutil.copy(fp, tempdir)
    with fitz.Document(temppdf) as doc:
        doc.del_xml_metadata()
        new_meta = doc.metadata or {}
        what, value = doc.xref_get_key(-1, "Info")  # /Info key in the trailer
        xref = int(value.replace("0 R", ""))

        if meta.name:
            new_meta["title"] = meta.name
            doc.xref_set_key(xref, "iscc_name", fitz.get_pdf_str(meta.name))
        if meta.description:
            new_meta["subject"] = meta.description
            doc.xref_set_key(xref, "iscc_description", fitz.get_pdf_str(meta.description))
        if meta.creator:
            new_meta["author"] = meta.creator
        if meta.keywords:
            new_meta["keywords"] = meta.keywords
        if meta.meta:
            doc.xref_set_key(xref, "iscc_meta", fitz.get_pdf_str(meta.meta))
        if meta.license:
            doc.xref_set_key(xref, "iscc_license", fitz.get_pdf_str(meta.license))
        if meta.acquire:
            doc.xref_set_key(xref, "iscc_acquire", fitz.get_pdf_str(meta.acquire))
        if meta.credit:
            doc.xref_set_key(xref, "iscc_credit", fitz.get_pdf_str(meta.credit))
        if meta.rights:
            doc.xref_set_key(xref, "iscc_rights", fitz.get_pdf_str(meta.rights))
        doc.set_metadata(new_meta)
        doc.saveIncr()
    return Path(temppdf)



"""Installer script for binary dependencies"""

import sys
from loguru import logger as log
import iscc_sdk as idk


def main():
    log.add(sys.stdout)
    log.info(f"Installing tools to {idk.dirs.user_data_dir}")
    idk.tools.install()
    log.info(f"FPCALC Version: {idk.tools.fpcalc_version_info()}")
    log.info(f"FFMPEG Version: {idk.tools.ffmpeg_version_info()}")

    sys.exit(0)


if __name__ == "__main__":
    main()



"""*Text handling functions*."""

import re
from html import unescape
from os.path import basename, splitext
from pathlib import Path
from typing import Generator, Union
from urllib.parse import urlparse

import bleach
from extractous import Extractor
from iscc_schema import IsccMeta
from PIL import Image
import xxhash
from loguru import logger as log
import iscc_sdk as idk
import iscc_lib as il

__all__ = [
    "text_meta_extract",
    "text_meta_embed",
    "text_extract",
    "text_features",
    "text_chunks",
    "text_name_from_uri",
    "text_thumbnail",
    "text_sanitize",
]
SCRIPT_RE = re.compile(r"<script[^>]*>.*?</script>", flags=re.DOTALL | re.IGNORECASE)
STYLE_RE = re.compile(r"<style[^>]*>.*?</style>", flags=re.DOTALL | re.IGNORECASE)

TEXT_META_MAP = {
    "iscc_name": "name",
    "iscc_description": "description",
    "iscc_meta": "meta",
    "iscc_license": "license",
    "iscc_acquire": "acquire",
    "iscc_credit": "credit",
    "iscc_rights": "rights",
    "custom:iscc_name": "name",
    "custom:iscc_description": "description",
    "custom:iscc_meta": "meta",
    "pdf:docinfo:title": "name",
    "pdf:docinfo:subject": "description",
    "pdf:docinfo:author": "creator",
    "pdf:docinfo:creator": "creator",
    "pdf:docinfo:keywords": "keywords",
    "dc:title": "name",
    "dc:description": "description",
    "dc:creator": "creator",
    "dc:rights": "rights",
    "meta:keyword": "keywords",
}


def text_meta_extract(fp):
    # type: (str|Path) -> dict
    """
    Extract metadata from text document file.

    :param fp: Filepath to text document file.
    :return: Metadata mapped to IsccMeta schema
    """
    fp = Path(fp)
    extractor = Extractor()
    # extractor = extractor.set_extract_string_max_length()
    result, meta = extractor.extract_file_to_string(fp.as_posix())
    mapped = dict()
    done = set()
    for tag, mapped_field in TEXT_META_MAP.items():
        if mapped_field in done:  # pragma nocover
            continue
        value = meta.get(tag)
        if value:
            if isinstance(value, list):
                value = ", ".join(value)
            value = idk.text_sanitize(value).strip()
            if value:
                log.debug(f"Mapping text metadata: {tag} -> {mapped_field} -> {value}")
                mapped[mapped_field] = value
                done.add(mapped_field)
    return mapped


def text_meta_embed(fp, meta):
    # type: (str|Path, IsccMeta) -> Path|None
    """
    Embed metadata into a copy of the text document.

    :param fp: Filepath to source text document file
    :param meta: Metadata to embed into text document
    :return: Filepath to the new file with updated metadata (None if no embedding supported)
    """
    fp = Path(fp)
    mt, _ = idk.mediatype_and_mode(fp)
    if mt == "application/pdf":
        return idk.pdf_meta_embed(fp, meta)
    if mt == "application/epub+zip":
        return idk.epub_meta_embed(fp, meta)
    if mt == "application/vnd.openxmlformats-officedocument.wordprocessingml.document":
        return idk.docx_meta_embed(fp, meta)
    return None


def text_extract(fp):
    # type: (str|Path) -> str
    """
    Extract plaintext from a text document.

    :param fp: Filepath to text document file.
    :return: Extracted plaintext
    """
    fp = Path(fp)
    if fp.suffix.lower() == ".pdf":
        return idk.pdf_text_extract(fp)
    extractor = Extractor()
    result, metadata = extractor.extract_file_to_string(fp.as_posix())
    text = result.strip()
    if not text:
        raise idk.IsccExtractionError(f"No text extracted from {fp.name}")
    return result


def text_features(text, **options):
    # type: (str) -> dict
    """
    Create granular simprints for text (minhashes over ngrams from cdc-chunks).
    Text should be normalized and cleaned before extracting text features.

    :param text: Normalized and cleaned plaintext.
    :param options: Keyword arguments forwarded to ``sdk_opts``:
        **byte_offsets** - Calculate offsets and sizes in UTF-8 bytes instead of chars. Default: False
    :returns dict: Dictionary with 'sizes', 'features', 'offsets', and 'contents'.
    """
    opts = idk.sdk_opts.override(options)
    sizes = []
    simprints = []
    offsets = []
    current_offset = 0
    for chunk in text_chunks(text, avg_size=idk.sdk_opts.text_avg_chunk_size):
        offsets.append(current_offset)
        ngrams = (
            "".join(chars)
            for chars in il.sliding_window(il.text_collapse(chunk), idk.core_opts.text_ngram_size)
        )
        features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
        minimum_hash_digest = il.alg_minhash_256(features)
        chunk_len = len(chunk.encode("utf-8")) if opts.byte_offsets else len(chunk)
        sizes.append(chunk_len)
        simprints.append(il.encode_base64(minimum_hash_digest))
        current_offset += chunk_len
    return dict(
        maintype="content",
        subtype="text",
        version=0,
        byte_offsets=opts.byte_offsets,
        simprints=simprints,
        offsets=offsets,
        sizes=sizes,
    )


def text_chunks(text, avg_size=idk.sdk_opts.text_avg_chunk_size):
    # type: (str, int) -> Generator[str, None, None]
    """
    Generates variable-sized text chunks (without leading BOM).

    :param text: Normalized plaintext
    :param avg_size: Targeted average size of text chunks in characters.
    :yields: Text chunks.
    """
    data = text.encode("utf-32-be")
    avg_size_bytes = avg_size * 4  # 4 bytes per character in utf-32-be
    for chunk_bytes in il.alg_cdc_chunks(data, utf32=True, avg_chunk_size=avg_size_bytes):
        yield chunk_bytes.decode("utf-32-be")


def text_name_from_uri(uri):
    # type: (Union[str, Path]) -> str
    """
    Extract `filename` part of an uri without file extension to be used as a fallback title for an
    asset if no title information can be acquired.

    :param uri: Url or file path
    :return: derived name (might be an empty string)
    """
    if isinstance(uri, Path):
        # Handle Path objects directly to avoid as_uri() issues with relative paths
        base = uri.name
    else:
        result = urlparse(uri)
        base = basename(result.path) if result.path else basename(result.netloc)
    name = splitext(base)[0]
    name = name.replace("-", " ")
    name = name.replace("_", " ")
    return name


def text_thumbnail(fp):
    # type: (str|Path) -> Image.Image|None
    """
    Create a thumbnail for a text document.

    :param fp: Filepath to a text document.
    :return: Thumbnail image as a PIL Image object
    """
    mt, _ = idk.mediatype_and_mode(fp)
    if mt == "application/pdf":
        return idk.pdf_thumbnail(fp)
    if mt == "application/epub+zip":
        return idk.epub_thumbnail(fp)
    return None


def text_sanitize(text):
    # type: (str) -> str
    """Sanitize text from untrusted sources (metadata extracted from assets)"""
    # Pre-process to remove script and style content
    text = SCRIPT_RE.sub("", text)
    text = STYLE_RE.sub("", text)

    # Sanitize with bleach - remove all HTML tags
    sanitized = bleach.clean(
        text,
        tags={},
        attributes={},
        strip=True,
        strip_comments=True,
    )

    # Decode HTML entities
    clean_text = unescape(sanitized)

    # Normalize whitespace
    clean_text = re.sub(r"\s+", " ", clean_text).strip()

    return clean_text



"""*Container format processing module*."""

from pathlib import Path
from typing import List, Dict, Callable, Optional
import iscc_sdk as idk

__all__ = [
    "process_container",
    "register_container_processor",
]

# Registry for container processors
_CONTAINER_PROCESSORS: Dict[str, Callable] = {}


def register_container_processor(mediatype: str, processor: Callable):
    """Register a processor function for a specific container mediatype."""
    _CONTAINER_PROCESSORS[mediatype] = processor


def process_container(fp, **options):
    # type: (str|Path, Any) -> Optional[List[idk.IsccMeta]]
    """
    Process embedded elements in a container file.

    :param fp: Filepath to container file
    :param options: Processing options
    :return: List of IsccMeta objects for embedded elements or None
    """
    fp = Path(fp)
    mediatype, _ = idk.mediatype_and_mode(fp)

    processor = _CONTAINER_PROCESSORS.get(mediatype)
    if processor:
        return processor(fp, **options)
    return None
